# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_config import cfg

from kuryr.lib import constants as const
from kuryr.lib import exceptions
from kuryr.tests.unit import base

from kuryr.lib.segmentation_type_drivers import vlan


class VlanSegmentationDriverTest(base.TestCase):
    """Unit tests for VLAN segmentation driver."""

    def setUp(self):
        super(VlanSegmentationDriverTest, self).setUp()
        cfg.CONF.binding.default_driver = 'kuryr.lib.binding.drivers.vlan'

    def test_allocate_segmentation_id(self):
        vlan_seg_driver = vlan.SegmentationDriver()
        allocated_ids = set([1, 2, 3])

        vlan_id = vlan_seg_driver.allocate_segmentation_id(allocated_ids)

        self.assertNotIn(vlan_id, vlan_seg_driver.available_local_vlans)
        self.assertNotIn(allocated_ids, vlan_seg_driver.available_local_vlans)

    def test_allocate_segmentation_id_only_1_available(self):
        vlan_seg_driver = vlan.SegmentationDriver()
        allocated_ids = set(range(const.MIN_VLAN_TAG,
                            const.MAX_VLAN_TAG + 1))
        allocated_ids.remove(const.MAX_VLAN_TAG)

        vlan_id = vlan_seg_driver.allocate_segmentation_id(allocated_ids)

        self.assertNotIn(vlan_id, vlan_seg_driver.available_local_vlans)
        self.assertNotIn(allocated_ids, vlan_seg_driver.available_local_vlans)
        self.assertEqual(vlan_id, const.MAX_VLAN_TAG)

    def test_allocate_segmentation_id_no_allocated_ids(self):
        vlan_seg_driver = vlan.SegmentationDriver()
        vlan_id = vlan_seg_driver.allocate_segmentation_id()
        self.assertNotIn(vlan_id, vlan_seg_driver.available_local_vlans)

    def test_allocate_segmentation_id_no_available_vlans(self):
        vlan_seg_driver = vlan.SegmentationDriver()
        allocated_ids = set(range(const.MIN_VLAN_TAG,
                            const.MAX_VLAN_TAG + 1))

        self.assertRaises(exceptions.SegmentationIdAllocationFailure,
                          vlan_seg_driver.allocate_segmentation_id,
                          allocated_ids)

    @mock.patch('random.choice')
    def test_allocate_segmentation_id_max_retries(self, mock_choice):
        mock_choice.side_effect = [1, 1, 1]
        vlan_seg_driver = vlan.SegmentationDriver()
        allocated_ids = set([1, 2, 3])

        self.assertRaises(exceptions.SegmentationIdAllocationFailure,
                          vlan_seg_driver.allocate_segmentation_id,
                          allocated_ids)
        self.assertEqual(len(mock_choice.mock_calls), 3)

    @mock.patch('random.choice')
    def test_allocate_segmentation_id_2_retries(self, mock_choice):
        vlan_seg_driver = vlan.SegmentationDriver()
        vlan_seg_driver.available_local_vlans = set(range(1, 10))
        allocated_ids = set([1, 2, 3])
        mock_choice.side_effect = [1, 1, 5]

        vlan_id = vlan_seg_driver.allocate_segmentation_id(allocated_ids)

        self.assertEqual(len(mock_choice.mock_calls), 3)
        self.assertEqual(vlan_id, 5)

    def test_release_segmentation_id(self):
        vlan_seg_driver = vlan.SegmentationDriver()
        vlan_seg_driver.available_local_vlans = set(range(1, 10))
        vlan_id = 20

        vlan_seg_driver.release_segmentation_id(vlan_id)

        self.assertIn(vlan_id, vlan_seg_driver.available_local_vlans)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

import ddt
from oslo_utils import uuidutils

import pyroute2.ipdb.interfaces

from kuryr.lib.binding.drivers import utils
from kuryr.lib import constants
from kuryr.tests.unit import base


@ddt.ddt
class BindingDriversUtilsTest(base.TestCase):
    """Unit tests for binding drivers utils"""

    def test_get_veth_pair_names(self):
        fake_neutron_port_id = uuidutils.generate_uuid()
        generated_ifname, generated_peer = utils.get_veth_pair_names(
            fake_neutron_port_id)

        namelen = constants.NIC_NAME_LEN
        ifname_postlen = namelen - len(constants.VETH_PREFIX)
        peer_postlen = namelen - len(constants.CONTAINER_VETH_PREFIX)

        self.assertEqual(namelen, len(generated_ifname))
        self.assertEqual(namelen, len(generated_peer))
        self.assertIn(constants.VETH_PREFIX, generated_ifname)
        self.assertIn(constants.CONTAINER_VETH_PREFIX, generated_peer)
        self.assertIn(fake_neutron_port_id[:ifname_postlen], generated_ifname)
        self.assertIn(fake_neutron_port_id[:peer_postlen], generated_peer)

    @ddt.data((False), (True))
    def test_is_up(self, interface_flag):
        fake_interface = {}
        if interface_flag:
            fake_interface['state'] = 'up'
            self.assertEqual(True, utils.is_up(fake_interface))
        else:
            self.assertEqual(False, utils.is_up(fake_interface))

    @ddt.data(
        (['10.10.10.11'], ['384ac9fc-eefa-4399-8d88-1181433e33b1'], False,
         None, None),
        (['10.10.10.11'], ['384ac9fc-eefa-4399-8d88-1181433e33b1'],
         True, None, None),
        (['10.10.10.11', '10.11.0.10'],
         ['384ac9fc-eefa-4399-8d88-1181433e33b1',
          '0a6eab28-9dc1-46c0-997c-cb9f66f6081f'],
         False, 1500, 'fa:16:3e:22:a3:3d'))
    @ddt.unpack
    @mock.patch.object(utils, 'is_up')
    def test__configure_container_iface(
            self, addrs, subnet_ids, already_up, mtu, mac, mock_is_up):
        subnets = [{
            'allocation_pools': [{'end': '10.11.0.254', 'start': '10.11.0.2'}],
            'cidr': '10.11.0.0/26',
            'created_at': '2016-09-27T07:55:12',
            'description': '',
            'dns_nameservers': [],
            'enable_dhcp': True,
            'gateway_ip': '10.11.0.1',
            'host_routes': [],
            'id': '0a6eab28-9dc1-46c0-997c-cb9f66f6081f',
            'ip_version': 4,
            'ipv6_address_mode': None,
            'ipv6_ra_mode': None,
            'name': 'subtest',
            'network_id': '90146ed2-c3ce-4001-866e-e97e513530a3',
            'revision': 2,
            'service_types': [],
            'subnetpool_id': None,
            'tenant_id': '0c0d1f46fa8d485d9534ea0e35f37bd3',
            'updated_at': '2016-09-27T07:55:12'
        }, {
            'allocation_pools': [{'end': '10.10.0.254', 'start': '10.10.0.2'}],
            'cidr': '10.10.0.0/24',
            'created_at': '2016-09-27T08:57:13',
            'description': '',
            'dns_nameservers': [],
            'enable_dhcp': True,
            'gateway_ip': '10.10.0.1',
            'host_routes': [],
            'id': '384ac9fc-eefa-4399-8d88-1181433e33b1',
            'ip_version': 4,
            'ipv6_address_mode': None,
            'ipv6_ra_mode': None,
            'name': '10.10.0.0/24',
            'network_id': 'bfb2f525-bedf-48ed-b125-102ee7920253',
            'revision': 2,
            'service_types': [],
            'subnetpool_id': None,
            'tenant_id': '51b66b97a12f42a990452967d2c555ac',
            'updated_at': '2016-09-27T08:57:13'}]

        fake_iface = mock.Mock(spec=pyroute2.ipdb.interfaces.Interface)
        _set_mtu = mock.Mock()
        _set_address = mock.Mock()
        fake_iface.attach_mock(_set_mtu, 'set_mtu')
        fake_iface.attach_mock(_set_address, 'set_address')
        mock_is_up.return_value = already_up

        fixed_ips = []
        for ip, subnet_id in zip(addrs, subnet_ids):
            fixed_ips.append({
                utils.IP_ADDRESS_KEY: ip,
                utils.SUBNET_ID_KEY: subnet_id})

        utils._configure_container_iface(
            fake_iface,
            subnets,
            fixed_ips,
            mtu=mtu,
            hwaddr=mac)

        subnets_prefix_by_id = dict(
            (subnet['id'], int(subnet['cidr'].split('/')[1])) for
            subnet in subnets)
        for ip, subnet_id in zip(addrs, subnet_ids):
            fake_iface.add_ip.assert_any_call(
                ip, subnets_prefix_by_id[subnet_id])

        if already_up:
            fake_iface.up.assert_not_called()
        else:
            fake_iface.up.assert_called_once()

        if mtu is None:
            fake_iface.set_mtu.assert_not_called()
        else:
            fake_iface.set_mtu.assert_called_with(mtu)

        if mac is None:
            fake_iface.set_address.assert_not_called()
        else:
            fake_iface.set_address.assert_called_with(mac)

    def test_get_ipdb(self):
        ip = utils.get_ipdb()
        self.assertEqual(ip, utils.get_ipdb())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import random

from kuryr.lib import constants as const
from kuryr.lib import exceptions

DEFAULT_MAX_RETRY_COUNT = 3


class SegmentationDriver(object):
    def __init__(self):
        self.available_local_vlans = set(range(const.MIN_VLAN_TAG,
                                         const.MAX_VLAN_TAG + 1))

    def allocate_segmentation_id(self, allocated_ids=set()):
        self.available_local_vlans.difference_update(allocated_ids)
        for i in range(DEFAULT_MAX_RETRY_COUNT):
            try:
                allocated = random.choice(list(self.available_local_vlans))
                self.available_local_vlans.remove(allocated)
                return allocated
            except IndexError:
                raise exceptions.SegmentationIdAllocationFailure(
                    'There are no vlan ids available.')
            except KeyError:
                # Other thread obtained the same vlan_id, so a new try is
                # needed
                continue
        raise exceptions.SegmentationIdAllocationFailure(
            'Max number of retries reached without '
            'finding an available vlan id.')

    def release_segmentation_id(self, id):
        self.available_local_vlans.add(id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import random
import socket

from keystoneauth1 import loading as ks_loading
from neutronclient.v2_0 import client
from oslo_config import cfg

from kuryr.lib import config as kuryr_config

DOCKER_NETNS_BASE = '/var/run/docker/netns'
PORT_POSTFIX = 'port'


def get_auth_plugin(conf_group):
    return ks_loading.load_auth_from_conf_options(
        cfg.CONF, conf_group)


def get_keystone_session(conf_group, auth_plugin):
    return ks_loading.load_session_from_conf_options(cfg.CONF,
                                                     conf_group,
                                                     auth=auth_plugin)


def get_neutron_client(*args, **kwargs):
    conf_group = kuryr_config.neutron_group.name
    auth_plugin = get_auth_plugin(conf_group)
    session = get_keystone_session(conf_group, auth_plugin)
    endpoint_type = getattr(getattr(cfg.CONF, conf_group), 'endpoint_type')
    region_name = getattr(getattr(cfg.CONF, conf_group), 'region_name')

    return client.Client(session=session,
                         auth=auth_plugin,
                         endpoint_type=endpoint_type,
                         region_name=region_name)


def get_hostname():
    """Returns the host name."""
    return socket.gethostname()


def get_neutron_subnetpool_name(subnet_cidr):
    """Returns a Neutron subnetpool name.

    :param subnet_cidr: The subnetpool allocation cidr
    :returns: the Neutron subnetpool_name name formatted appropriately
    """
    name_prefix = cfg.CONF.subnetpool_name_prefix
    return '-'.join([name_prefix, subnet_cidr])


def get_dict_format_fixed_ips_from_kv_format(fixed_ips):
    """Returns fixed_ips in dict format.

    :param fixed_ips: Format that neutron client expects for list_ports ex,
                      ['subnet_id=5083bda8-1b7c-4625-97f3-1d4c33bfeea8',
                       'ip_address=192.168.1.2']
    :returns: normal dict form,
              [{'subnet_id': '5083bda8-1b7c-4625-97f3-1d4c33bfeea8',
                'ip_address': '192.168.1.2'}]
    """
    new_fixed_ips = []
    for fixed_ip in fixed_ips:
        if 'subnet_id' == fixed_ip.split('=')[0]:
            subnet_id = fixed_ip.split('=')[1]
        else:
            ip = fixed_ip.split('=')[1]
            new_fixed_ips.append(
                {
                    'subnet_id': subnet_id,
                    'ip_address': ip
                }
            )
    return new_fixed_ips


def getrandbits(bit_size=256):
    return str(random.getrandbits(bit_size)).encode('utf-8')


def get_hash(bit_size=256):
    return hashlib.sha256(getrandbits(bit_size=bit_size)).hexdigest()


def string_mappings(mapping_list):
    """Make a string out of the mapping list"""
    details = ''
    if mapping_list:
        details = '"' + str(mapping_list) + '"'
        return details


def get_random_string(length):
    """Get a random hex string of the specified length."""

    return "{0:0{1}x}".format(random.getrandbits(length * 4), length)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class KuryrException(Exception):
    """Default Kuryr exception"""


class BindingFailure(KuryrException):
    """Exception represents the binding is failed.

    This exception is thrown when the executable script for the binding is
    failed and Kuryr can't proceed further.
    """


class BindingNotSupportedFailure(KuryrException):
    """Exception represents the vif type binding not support.

    This exception is thrown when the executable script for the binding does
    not exist and Kuryr can't proceed further.
    """


class DuplicatedResourceException(KuryrException):
    """Exception represents there're multiple resources for the ID.

    For example, this exception is thrown when you query the Neutron resource
    associated with the ID and you get multiple resources.
    """


class GatewayConflictFailure(KuryrException):
    """Exception represents gateway ip is conflict.

    This exception is thrown when request gateway ip is conflict with the
    gateway ip in existed network.
    """


class MandatoryApiMissing(KuryrException):
    """Exception represents that mandatory api is not found.

    For example, this exception is thrown when expected neutron
    extension(subnetpools) APIs are not found.
    """


class NoResourceException(KuryrException):
    """Exception represents there's no resource for the given query.

    This exception is thrown when you query the Neutron resource associated
    with the given query and you get none of them actually.
    """


class InactiveResourceException(KuryrException):
    """Exception represents the resource for the given query is not active.

    This exception is thrown when you query the Neutron resource associated
    with the given query and you get the status of the resource as something
    other than ACTIVE.
    """


class VethCreationFailure(KuryrException):
    """Exception represents the veth pair creation is failed.

    This exception is thrown when the veth pair is not created appropriately
    and Kuryr can't proceed the binding further.
    """


class VethDeletionFailure(KuryrException):
    """Exception represents the veth pair deletion is failed.

    This exception is thrown when the veth pair is not deleted appropriately
    and Kuryr can't proceed the unbinding further.
    """


class ExportPortFailure(KuryrException):
    """Exception represents setting up exported port is failed.

    This exception is thrown when performing Neutron security group failed
    for an exported port and Kuryr can't proceed the expose further.
    """


class SegmentationIdAllocationFailure(KuryrException):
    """Exception represents when segmentation id could not be allocated.

    This exception is thrown when the segmentaion id for the isolation of
    container traffic could not be allocated and Kuryr can't proceed further.
    """


class SegmentationDriverBindingDriverCompatibilityFailure(KuryrException):
    """Exception represents when no segmentation type driver is loaded.

    This exception is thrown when configured binding driver does not have
    a supporting segmentation type driver.
    """


class AddressInUseException(KuryrException):
    """Exception represents the requested ip address is used.

    This exception is thrown when a specific address is requested and the
    requested ip address already exists and is used.
    """


class DriverNotEnabledException(KuryrException):
    """Exception represents the binding driver is not enabled.

    This exception is thrown when kuryr tries to load a specific binding driver
    but the driver is not enabled.
    """


class ConflictConfigOption(KuryrException):
    """Exception represents when the conflict config happens.

    The existed option conflicts with the given option.
    """



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from oslo_utils import uuidutils

from kuryr.lib.binding.drivers import hw_veb
from kuryr.lib import constants
from kuryr.lib import utils
from kuryr.tests.unit import base


mock_create = mock.MagicMock()
mock_interface = mock.MagicMock()


class TestHwVebDriver(base.TestCase):
    """Unit tests for hw_veb driver"""

    @mock.patch('oslo_concurrency.processutils.execute',
                return_value=('fake_stdout', 'fake_stderr'))
    def test_port_bind(self, mock_execute):
        fake_docker_endpoint_id = utils.get_hash()
        fake_docker_network_id = utils.get_hash()
        fake_port_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_vlan_id = 100
        fake_vif_details = {constants.VIF_DETAILS_VLAN_KEY: fake_vlan_id}
        fake_vif_type = "ovs"
        fake_port = self._get_fake_port(
            fake_docker_endpoint_id, fake_docker_network_id,
            fake_port_id, constants.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            vif_details=fake_vif_details, vif_type=fake_vif_type)
        fake_subnets = self._get_fake_subnets(
            fake_docker_endpoint_id, fake_docker_network_id,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id)
        fake_network = self._get_fake_networks(fake_docker_network_id)
        fake_pf_ifname = 'eth13'
        fake_vf_num = 1

        hw_veb.port_bind(fake_docker_endpoint_id,
                         fake_port['port'],
                         fake_subnets['subnets'],
                         fake_network['networks'][0],
                         pf_ifname=fake_pf_ifname,
                         vf_num=fake_vf_num)

        mock_execute.assert_called_once_with(
            'ip', 'link', 'set', fake_pf_ifname,
            'vf', fake_vf_num,
            'mac', fake_port['port']['mac_address'],
            'vlan', fake_vlan_id,
            check_exit_code=[0, 2, 254])

    @mock.patch('oslo_concurrency.processutils.execute',
                return_value=('fake_stdout', 'fake_stderr'))
    def test_port_unbind(self, mock_execute):
        fake_docker_endpoint_id = utils.get_hash()
        fake_docker_network_id = utils.get_hash()
        fake_port_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_vif_type = "ovs"
        fake_port = self._get_fake_port(
            fake_docker_endpoint_id, fake_docker_network_id,
            fake_port_id, constants.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            vif_type=fake_vif_type)
        fake_pf_ifname = 'eth13'
        fake_vf_num = 1
        hw_veb.port_unbind(fake_docker_endpoint_id, fake_port['port'],
                           pf_ifname=fake_pf_ifname,
                           vf_num=fake_vf_num)
        mock_execute.assert_called_once()
        mock_execute.assert_called_once_with(
            'ip', 'link', 'set', fake_pf_ifname,
            'vf', fake_vf_num,
            'mac', fake_port['port']['mac_address'],
            'vlan', 0,
            check_exit_code=[0, 2, 254])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg
from oslo_utils import importutils

from kuryr.lib import exceptions as ex

BASE_PATH = 'kuryr.lib.segmentation_type_drivers'

_driver = ""


def _get_driver():
    global _driver
    if not _driver:
        driver_name = cfg.CONF.binding.default_driver.rsplit('.', 1)[1]

        # REVISIT(vikasc): Need to remove this if check
        if driver_name == 'vlan':
            seg_driver_path = '.'.join([BASE_PATH, driver_name])
            segmentation_driver = importutils.import_module(seg_driver_path)
            _driver = segmentation_driver.SegmentationDriver()
    return _driver


def allocate_segmentation_id(allocated_ids=set()):
    """Allocates a segmentation ID."""
    driver = _get_driver()
    try:
        id = driver.allocate_segmentation_id(allocated_ids)
    except NameError:
        raise ex.SegmentationDriverBindingDriverCompatibilityFailure
    return id


def release_segmentation_id(id):
    """Releases the segmentation ID."""
    driver = _get_driver()
    try:
        driver.release_segmentation_id(id)
    except NameError:
        raise ex.SegmentationDriverBindingDriverCompatibilityFailure



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Kuryr
"""

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from kuryr.lib._i18n import _


core_opts = [
    cfg.StrOpt('bindir',
               default='/usr/libexec/kuryr',
               help=_('Directory for Kuryr vif binding executables.')),
    cfg.StrOpt('subnetpool_name_prefix',
               default='kuryrPool',
               help=_('Neutron subnetpool name will be prefixed by this.')),
    cfg.StrOpt('deployment_type',
               default='baremetal',
               help=_("baremetal or nested-containers are the supported"
                      " values.")),
]

neutron_group = cfg.OptGroup(
    'neutron',
    title='Neutron Options',
    help=_('Configuration options for OpenStack Neutron'))

neutron_opts = [
    cfg.StrOpt('enable_dhcp',
               default='True',
               help=_('Enable or Disable dhcp for neutron subnets.')),
    cfg.StrOpt('default_subnetpool_v4',
               default='kuryr',
               help=_('Name of default subnetpool version 4')),
    cfg.StrOpt('default_subnetpool_v6',
               default='kuryr6',
               help=_('Name of default subnetpool version 6')),
    cfg.BoolOpt('vif_plugging_is_fatal',
                default=False,
                help=_("Whether a plugging operation is failed if the port "
                       "to plug does not become active")),
    cfg.IntOpt('vif_plugging_timeout',
               default=0,
               help=_("Seconds to wait for port to become active")),
    cfg.StrOpt('region_name',
               default=None,
               help=_('Region name of the neturon endpoint to use.')),
    cfg.StrOpt('endpoint_type',
               default='public',
               choices=['public', 'admin', 'internal'],
               help=_('Type of the neutron endpoint to use. This endpoint '
                      'will be looked up in the keystone catalog and should '
                      'be one of public, internal or admin.')),
]

binding_opts = [
    cfg.StrOpt('veth_dst_prefix',
               default='eth',
               help=_('The name prefix of the veth endpoint put inside the '
                      'container.')),
    cfg.StrOpt('default_driver',
               default='kuryr.lib.binding.drivers.veth',
               deprecated_name='driver',
               help=_('Driver to use for binding and unbinding ports.')),
    cfg.ListOpt('enabled_drivers',
                default=['kuryr.lib.binding.drivers.veth'],
                help=_('Drivers to use for binding and unbinding ports.')),
    cfg.StrOpt('link_iface',
               default='',
               help=_('Specifies the name of the Nova instance interface to '
                      'link the virtual devices to (only applicable to some '
                      'binding drivers.')),
]

binding_group = cfg.OptGroup(
    'binding',
    title='binding options',
    help=_('Configuration options for container interface binding.'))


def register_keystoneauth_opts(conf, conf_group):
    ks_loading.register_session_conf_options(conf, conf_group)
    ks_loading.register_auth_conf_options(conf, conf_group)


def register_neutron_opts(conf):
    conf.register_group(neutron_group)
    conf.register_opts(neutron_opts, group=neutron_group)
    register_keystoneauth_opts(conf, neutron_group.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


PORT_STATUS_ACTIVE = 'ACTIVE'
PORT_STATUS_DOWN = 'DOWN'

DEVICE_OWNER = 'compute:kuryr'
NIC_NAME_LEN = 14
VETH_PREFIX = 'tap'
CONTAINER_VETH_PREFIX = 't_c'

# For VLAN type segmentation
MIN_VLAN_TAG = 1
MAX_VLAN_TAG = 4094

BINDING_SUBCOMMAND = 'bind'
DEFAULT_NETWORK_MTU = 1500
FALLBACK_VIF_TYPE = 'unbound'
UNBINDING_SUBCOMMAND = 'unbind'
VIF_DETAILS_KEY = 'binding:vif_details'
VIF_DETAILS_VLAN_KEY = 'vlan'
VIF_TYPE_KEY = 'binding:vif_type'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

__all__ = [
    'list_kuryr_opts',
]

import copy
import itertools
import operator

from keystoneauth1 import loading as ks_loading
from oslo_log import _options

from kuryr.lib import config


ENABLED_AUTH_PLUGINS = ('password', 'v2password', 'v2token', 'v3password',
                        'v3token')

_core_opts_with_logging = config.core_opts
_core_opts_with_logging += _options.common_cli_opts
_core_opts_with_logging += _options.logging_cli_opts
_core_opts_with_logging += _options.generic_log_opts

_kuryr_opts = [
    (None, list(itertools.chain(_core_opts_with_logging))),
]


def get_keystoneauth_conf_options():
    opt_list = []
    opt_list.insert(0, ks_loading.get_auth_common_conf_options()[0])
    opt_list += ks_loading.get_session_conf_options()
    # NOTE(apuimedo): There are a lot of auth plugins, we just generate the
    # config options for a few common ones
    for name in ENABLED_AUTH_PLUGINS:
        for plugin_option in ks_loading.get_auth_plugin_conf_options(name):
            if all(option.name != plugin_option.name for option in opt_list):
                opt_list.append(plugin_option)
    return opt_list


def list_neutron_opts():
    opt_list = copy.deepcopy(config.neutron_opts)
    opt_list += get_keystoneauth_conf_options()
    opt_list.sort(key=operator.attrgetter('name'))
    return [(config.neutron_group, opt_list)]


def list_kuryr_opts():
    """Return a list of oslo_config options available in Kuryr service.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    This function is also discoverable via the 'kuryr' entry point under
    the 'oslo_config.opts' namespace.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users by Kuryr.

    :returns: a list of (group_name, opts) tuples
    """

    opts = [(k, copy.deepcopy(o)) for k, o in _kuryr_opts]
    opts += list_neutron_opts()
    opts += [(config.binding_group, config.binding_opts)]
    return opts



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import ipaddress

import pyroute2

from kuryr.lib import constants


_IPDB_CACHE = None

FIXED_IP_KEY = 'fixed_ips'
IP_ADDRESS_KEY = 'ip_address'
MAC_ADDRESS_KEY = 'mac_address'
SUBNET_ID_KEY = 'subnet_id'


def get_veth_pair_names(port_id):
    ifname = constants.VETH_PREFIX + port_id
    ifname = ifname[:constants.NIC_NAME_LEN]
    peer_name = constants.CONTAINER_VETH_PREFIX + port_id
    peer_name = peer_name[:constants.NIC_NAME_LEN]
    return ifname, peer_name


def get_ipdb():
    """Returns the already cached or a newly created IPDB instance.

    IPDB reads the Linux specific file when it's instantiated. This behaviour
    prevents Mac OSX users from running unit tests. This function makes the
    loading IPDB lazyily and therefore it can be mocked after the import of
    modules that import this module.

    :returns: The already cached or newly created ``pyroute2.IPDB`` instance
    """
    global _IPDB_CACHE
    if not _IPDB_CACHE:
        _IPDB_CACHE = pyroute2.IPDB()
    return _IPDB_CACHE


def get_mtu_from_network(network=None):
    """Get Maximum Transfer Unit from neutron network.

    :parm network: neutron network
    :returns: mtu on the neutron network
    """
    if network is None:
        mtu = constants.DEFAULT_NETWORK_MTU
    else:
        mtu = network.get('mtu', constants.DEFAULT_NETWORK_MTU)
    return mtu


def remove_device(ifname):
    """Removes the device with name ifname.

    :param ifname: the name of the device to remove
    :returns: the index the device identified by ifname had if it
              exists, otherwise None
    :raises: pyroute2.NetlinkError
    """
    ip = get_ipdb()

    dev_index = ip.interfaces.get(ifname, {}).get('index', None)

    if dev_index:
        with ip.interfaces[ifname] as iface:
            iface.remove()

    return dev_index


def is_up(interface):
    return interface.get('state', '') == 'up'


def _configure_container_iface(iface, subnets, fixed_ips, mtu=None,
                               hwaddr=None):
    """Configures the interface that is placed in the container net ns

    :param iface:       the pyroute IPDB interface object to configure
    :param subnets:     an iterable of all the Neutron subnets which the
                        endpoint is trying to join
    :param fixed_ips:   an iterable of fixed IPs to be set for the iface
    :param mtu:         Maximum Transfer Unit to set for the iface
    :param hwaddr:      Hardware address to set for the iface
    """
    subnets_dict = {subnet['id']: subnet for subnet in subnets}
    # We assume containers always work with fixed ips, dhcp does not really
    # make a lot of sense
    for fixed_ip in fixed_ips:
        if IP_ADDRESS_KEY in fixed_ip and (SUBNET_ID_KEY in fixed_ip):
            subnet_id = fixed_ip[SUBNET_ID_KEY]
            subnet = subnets_dict[subnet_id]
            cidr = ipaddress.ip_network(str(subnet['cidr']))
            iface.add_ip(fixed_ip[IP_ADDRESS_KEY], cidr.prefixlen)
    if mtu is not None:
        iface.set_mtu(mtu)
    if hwaddr is not None:
        iface.set_address(hwaddr)
    if not is_up(iface):
        iface.up()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_utils import importutils

from kuryr.lib import binding
from kuryr.lib import exceptions
from kuryr.tests.unit import base


class TestBinding(base.TestCase):
    """Unit tests for binding module"""

    def test__verify_driver(self):
        cfg.CONF.set_override('enabled_drivers',
                              ['kuryr.lib.binding.drivers.veth'],
                              group='binding')
        driver = importutils.import_module('kuryr.lib.binding.drivers.veth')
        binding._verify_driver(driver)  # assert no exception raise
        driver = importutils.import_module('kuryr.lib.binding.drivers.vlan')
        self.assertRaises(exceptions.DriverNotEnabledException,
                          binding._verify_driver, driver)
